import os
import ifcopenshell
import ifcopenshell.api
import ifcopenshell.util
import ifcopenshell.guid
import ifcopenshell.util.element
import logging

# Check if there is an IfcProject,
# Check if there is an IfcSite aggregated by that IfcProject
# Check what kind of facility is aggregated by the IfcSite:
      # If the facility is a IfcBridge:
        # use it as main container of the IfcBridgeParts
      # If the facility is a IfcBuilding:
        #copy the main attributes and property sets
        #create a new IfcBridge, translate the attributes and the possible psets into the new IfcBridge

# For each element -> if the corresponding IfcBridgePart does not exist:
    # Create an IfcBridgePart inside the IfcBridge
    # Else add a 
# Into that BridgePart create 

def traverse_ifc_bridge_aggregation(ifc_model: ifcopenshell.file):
    """
    Traverses the IFC model to replace IfcBuilding with IfcBridge, copying psets and attributes.
    """
    try:
        
        ifc_facility = get_ifc_facility(ifc_model)
        if ifc_facility is None:
            logging.error("No valid facility found in the model.")
            return

        elif ifc_facility.is_a("IfcBridge"):
            logging.info("Model already contains an IfcBridge. No action taken.")
            return ifc_facility

        elif ifc_facility.is_a("IfcBuilding"):
            building_attributes = {
                "Name": ifc_facility.Name,
                "Description": ifc_facility.Description,
                "ObjectType": ifc_facility.ObjectType
            }
            
            new_bridge = create_ifc_bridge(
                ifc_model,
                bridge_attributes=building_attributes,
                ifc_building=ifc_facility
            )
            bridge_psets = ifcopenshell.util.element.get_psets(new_bridge)
            print(bridge_psets)
            logging.info(f'Bridge psets: {bridge_psets}')
            
            return new_bridge
        else:
            logging.warning(f"Unhandled facility type: {ifc_facility.is_a()}")
    except Exception as e:
        logging.error(f"Error in traverse_ifc_aggregation: {e}")
        return None

def element_has_aggregation(ifc_element):
    '''
    Auxiliar function to check aggregation of 
    '''
    site_aggregates = ifc_element.IsDecomposedBy
    if not site_aggregates:
        return False
    for rel in site_aggregates:
        if len(rel.RelatedObjects)> 0:
            return True

def get_ifc_site(ifc_model: ifcopenshell.file):
    """
    Retrieves the primary IfcSite from the IFC model.
    Assumes there's only one site with aggregations.
    """
    sites = ifc_model.by_type("IfcSite")
    sites_with_aggregations = [site for site in sites if element_has_aggregation(site)]
    
    if len(sites_with_aggregations) == 1:
        logging.info("Primary IfcSite found.")
        return sites_with_aggregations[0]
    elif len(sites_with_aggregations) > 1:
        logging.warning("Multiple IfcSites with aggregations found. Using the first one.")
        return sites_with_aggregations[0]
    else:
        logging.error("No IfcSite with aggregation found")
        return None

def get_ifc_facility(ifc_model: ifcopenshell.file):
    '''
    Gets the main facility from a model if the facility is a IfcBridge or IfcBuilding
    '''
    projects = ifc_model.by_type("IfcProject")
    if len(projects)==1:
        project = projects[0]
        project_aggregates= project.IsDecomposedBy

        ifc_sites = []
        for rel in project_aggregates:
            for obj in rel.RelatedObjects:
                if obj.is_a("IfcSite"):
                    ifc_sites.append(obj)
        
        ifc_sites_with_aggregations = []
        for site in ifc_sites:
            if element_has_aggregation(site):
                ifc_sites_with_aggregations.append(site)
        
        if len(ifc_sites_with_aggregations) == 1:
            model_site = ifc_sites_with_aggregations[0]
            site_aggregates = model_site.IsDecomposedBy
            
            ifc_facilities = []
            for rel in site_aggregates:
                for obj in rel.RelatedObjects:
                    if obj.is_a("IfcBridge") or obj.is_a("IfcBuilding"):
                        return obj
        else:
            print("Not valid Model")
            return None
    elif not project or len(projects)>1:
        raise ValueError("Not valid Model, more than one IfcProjects detected in file")
    return None

def create_ifc_owner_history(ifc_model: ifcopenshell.file):
    """
    Creates a default IfcOwnerHistory if none exists in the model.
    """
    try:
        # Create IfcPerson
        person = ifc_model.create_entity("IfcPerson", Identification="User", FamilyName="User")
        ifc_model.add(person)

        # Create IfcOrganization
        organization = ifc_model.create_entity("IfcOrganization", Name="Default Organization")
        ifc_model.add(organization)

        # Create IfcPersonAndOrganization
        person_and_org = ifc_model.create_entity(
            "IfcPersonAndOrganization",
            Person=person,
            Organization=organization
        )
        ifc_model.add(person_and_org)

        # Create IfcApplication
        application = ifc_model.create_entity(
            "IfcApplication",
            ApplicationDeveloper=organization,
            Version="1.0",
            ApplicationFullName="IfcOpenShell",
            ApplicationIdentifier="IfcOpenShell v0.6"
        )
        ifc_model.add(application)

        # Create IfcOwnerHistory
        owner_history = ifc_model.create_entity(
            "IfcOwnerHistory",
            OwningUser=person_and_org,
            OwningApplication=application,
            State=None,
            ChangeAction="ADDED",
            LastModifiedDate=None
        )
        ifc_model.add(owner_history)

        logging.info("Created default IfcOwnerHistory.")
        return owner_history
    except Exception as e:
        logging.error(f"Error creating IfcOwnerHistory: {e}")
        raise

def extract_property_value(prop_detail):
    """
    Extracts the value from an IfcProperty object based on its type.
    """
    try:
        if prop_detail.is_a("IfcPropertySingleValue"):
            # For single value properties, return the wrapped value
            return prop_detail.NominalValue.wrappedValue
        elif prop_detail.is_a("IfcPropertyEnumeratedValue"):
            # For enumerated values, return a list of enumeration values
            return [enum_value.WrappedValue for enum_value in prop_detail.EnumerationValues]
        elif prop_detail.is_a("IfcPropertyBoundedValue"):
            # For bounded values, return a dictionary with lower and upper bounds
            return {
                "LowerBoundValue": prop_detail.LowerBoundValue.wrappedValue,
                "UpperBoundValue": prop_detail.UpperBoundValue.wrappedValue
            }
        elif prop_detail.is_a("IfcPropertyListValue"):
            # For list values, return a list of wrapped values
            return [val.wrappedValue for val in prop_detail.ListValues]
        else:
            logging.warning(f"Unhandled property type: {prop_detail.is_a()}")
            return None
    except AttributeError as e:
        logging.error(f"Error extracting property value: {e}")
        return None

def create_ifc_bridge(ifc_model: ifcopenshell.file, bridge_attributes=None, ifc_building=None):
    """
    Creates an IfcBridge entity with the specified attributes and copies psets from the IfcBuilding.
    """
    try:
        # Retrieve or create OwnerHistory
        owner_histories = ifc_model.by_type("IfcOwnerHistory")
        if not owner_histories:
            owner_history = create_ifc_owner_history(ifc_model)
        else:
            owner_history = owner_histories[0]
            logging.info("Using existing IfcOwnerHistory.")

        # Get bridge attributes or set defaults
        attributes = bridge_attributes or {}
        bridge_name = attributes.get("Name", "Example Bridge")
        bridge_description = attributes.get("Description", "A sample IfcBridgeEntity")
        bridge_object_type = attributes.get("ObjectType", "Infrastructure")

        # Create the IfcBridge entity
        bridge = ifc_model.create_entity(
            "IfcBridge",
            GlobalId=ifcopenshell.guid.new(),
            OwnerHistory=owner_history,
            Name=bridge_name,
            Description=bridge_description,
            ObjectType=bridge_object_type
        )
        ifc_model.add(bridge)
        logging.info(f"Created IfcBridge with GlobalId {bridge.GlobalId}.")

        # Copy property sets from IfcBuilding to IfcBridge
        if ifc_building:
            psets = ifcopenshell.util.element.get_psets(ifc_building)
            logging.info(f"Found {len(psets)} property sets in IfcBuilding.")

            for pset_name, properties in psets.items():
                # Rename the pset if necessary
                new_pset_name = "Pset_BridgeCommon" if pset_name == "Pset_BuildingCommon" else pset_name

                # Create a new IfcPropertySet
                pset = ifc_model.create_entity(
                    "IfcPropertySet",
                    GlobalId=ifcopenshell.guid.new(),
                    OwnerHistory=owner_history,
                    Name=new_pset_name
                )

                # Create properties within the pset
                has_properties = []
                for prop_name, prop_value in properties.items():
                    # Determine the type of property based on the value
                    if isinstance(prop_value, str):
                        nominal_value = ifc_model.create_entity("IfcText", prop_value)
                        prop = ifc_model.create_entity(
                            "IfcPropertySingleValue",
                            Name=prop_name,
                            NominalValue=nominal_value
                        )
                    elif isinstance(prop_value, float):
                        nominal_value = ifc_model.create_entity("IfcReal", prop_value)
                        prop = ifc_model.create_entity(
                            "IfcPropertySingleValue",
                            Name=prop_name,
                            NominalValue=nominal_value
                        )
                    elif isinstance(prop_value, int):
                        nominal_value = ifc_model.create_entity("IfcInteger", prop_value)
                        prop = ifc_model.create_entity(
                            "IfcPropertySingleValue",
                            Name=prop_name,
                            NominalValue=nominal_value
                        )
                    elif isinstance(prop_value, list):
                        # Handle list values as IfcPropertyListValue
                        list_values = []
                        for val in prop_value:
                            if isinstance(val, float):
                                list_values.append(ifc_model.create_entity("IfcReal", val))
                            elif isinstance(val, int):
                                list_values.append(ifc_model.create_entity("IfcInteger", val))
                            elif isinstance(val, str):
                                list_values.append(ifc_model.create_entity("IfcText", val))
                            else:
                                logging.warning(f"Unsupported list item type for '{prop_name}': {type(val)}")
                        if list_values:
                            list_entity = ifc_model.create_entity(
                                "IfcPropertyListValue",
                                ListValues=list_values
                            )
                            prop = ifc_model.create_entity(
                                "IfcPropertySingleValue",
                                Name=prop_name,
                                NominalValue=list_entity
                            )
                        else:
                            logging.warning(f"No valid list items for property '{prop_name}'. Skipping.")
                            continue
                    else:
                        logging.warning(f"Unsupported property type for '{prop_name}': {type(prop_value)}")
                        continue

                    has_properties.append(prop)

                # Assign properties to the pset
                pset.HasProperties = has_properties
                ifc_model.add(pset)
                logging.info(f"Created property set '{new_pset_name}' with {len(has_properties)} properties.")

                # Associate the pset with the bridge
                rel_defines = ifc_model.create_entity(
                    "IfcRelDefinesByProperties",
                    GlobalId=ifcopenshell.guid.new(),
                    OwnerHistory=owner_history,
                    RelatingPropertyDefinition=pset,
                    RelatedObjects=[bridge]
                )
                ifc_model.add(rel_defines)
                logging.info(f"Associated property set '{new_pset_name}' with IfcBridge.")

        # Integrate the bridge into the spatial structure
        ifc_site = get_ifc_site(ifc_model)
        if ifc_site:
            rel_aggregates = ifc_model.create_entity(
                "IfcRelAggregates",
                GlobalId=ifcopenshell.guid.new(),
                OwnerHistory=owner_history,
                Name=None,
                Description=None,
                RelatingObject=ifc_site,
                RelatedObjects=[bridge]
            )
            ifc_model.add(rel_aggregates)
            logging.info("IfcBridge has been integrated into the spatial structure under IfcSite.")

        return bridge
    except Exception as e:
        print(e)

def copy_property_sets(ifc_model: ifcopenshell.file, source_entity, target_entity):
    """
    Copies property sets from a source entity to a target entity.
    """
    rels = ifc_model.get_inverse(source_entity)
    for rel in rels:
        if rel.is_a("IfcRelDefinesByProperties"):
            property_definition = rel.RelatingPropertyDefinition
            if property_definition.is_a("IfcPropertySet"):
                # Create new IfcPropertySet for the target entity
                new_property_set = ifc_model.create_entity(
                    "IfcPropertySet",
                    GlobalId=ifcopenshell.guid.new(),
                    OwnerHistory=property_definition.OwnerHistory,
                    Name=property_definition.Name,
                    Description=property_definition.Description
                )

                # Clone properties
                has_properties = []
                for prop in property_definition.HasProperties:
                    # Clone each property based on its type
                    if prop.is_a("IfcPropertySingleValue"):
                        cloned_prop = ifc_model.create_entity(
                            "IfcPropertySingleValue",
                            Name=prop.Name,
                            NominalValue=prop.NominalValue
                        )
                    elif prop.is_a("IfcPropertyEnumeratedValue"):
                        cloned_prop = ifc_model.create_entity(
                            "IfcPropertyEnumeratedValue",
                            Name=prop.Name,
                            EnumerationValues=prop.EnumerationValues
                        )
                    else:
                        logging.warning(f"Unsupported property type: {prop.is_a()}")
                        continue
                    has_properties.append(cloned_prop)

                new_property_set.HasProperties = has_properties
                ifc_model.add(new_property_set)
                logging.info(f"Copied property set '{property_definition.Name}' to target entity.")

                # Associate the new property set with the target entity
                new_rel = ifc_model.create_entity(
                    "IfcRelDefinesByProperties",
                    GlobalId=ifcopenshell.guid.new(),
                    OwnerHistory=property_definition.OwnerHistory,
                    RelatingPropertyDefinition=new_property_set,
                    RelatedObjects=[target_entity]
                )
                ifc_model.add(new_rel)
                logging.info(f"Associated property set '{property_definition.Name}' with target entity.")


def main(input_ifc_path: str, output_ifc_path: str):
    """
    Main function to execute the IFC modification.
    """
    try:
        # Open the IFC file
        ifc_model = ifcopenshell.open(input_ifc_path)
        logging.info(f"Opened IFC file: {input_ifc_path}")

        # Traverse and modify the IFC model
        traverse_ifc_bridge_aggregation(ifc_model)

        # Write the modified IFC model to a new file
        ifc_model.write(output_ifc_path)
        logging.info(f"Modified IFC file written to: {output_ifc_path}")
    except Exception as e:
        logging.error(f"An error occurred during processing: {e}")

if __name__=="__main__":
    ifc_paths = [
        r"C:\00_TECNALIA\01_PROJECTS\TCRINI\00_CODE\RemapBridges\public\ifc_bridges\model01_girder.ifc",
        r"C:\00_TECNALIA\01_PROJECTS\TCRINI\00_CODE\RemapBridges\public\ifc_bridges\model02_tablero.ifc",
        r"C:\00_TECNALIA\01_PROJECTS\TCRINI\00_CODE\RemapBridges\public\ifc_bridges\model03_zubiot.ifc",
        r"C:\00_TECNALIA\01_PROJECTS\TCRINI\00_CODE\RemapBridges\public\ifc_bridges\model04_kit.ifc",
        r"C:\00_TECNALIA\01_PROJECTS\TCRINI\00_CODE\RemapBridges\public\ifc_bridges\model05_berriz-IFC4-IFC4X3.ifc",
        r"C:\00_TECNALIA\01_PROJECTS\TCRINI\00_CODE\RemapBridges\public\ifc_bridges\model06_PrecastAndCastConcrete.ifc",
        r"C:\00_TECNALIA\01_PROJECTS\TCRINI\00_CODE\RemapBridges\public\ifc_bridges\Viadotto Acerno.ifc"
    ]
    input_file = ifc_paths[4]
    output_path = r"C:\00_TECNALIA\01_PROJECTS\TCRINI\00_CODE\RemapBridges\public\ifc_bridge_test_bridge\output.ifc"
    main(input_file,output_path)
