import multiprocessing
import numpy as np
import time 

import ifcopenshell
import ifcopenshell.geom
import ifcopenshell.util.element
import ifcopenshell.util.shape

from pprint import pprint

from ifc_bridge_mappings import BRIDGE_PARTS

class IfcBridgeElement:
    def __init__(self, IFC_ELEMENT, settings):
        self.IFC_ELEMENT = IFC_ELEMENT
        self.SETTINGS = settings
        self.SHAPE = None
        self.GEOMETRY = None
        self.HAS_REPRESENTATION = self.get_representation()
        
        self.ID_GUID = IFC_ELEMENT.GlobalId
        self.SEMA_CLASS = IFC_ELEMENT.is_a()
        self.SEMA_PTYPE = None
        
        self.GEOM_VERTICES = None
        self.GEOM_CORNERS = None
        self.GEOM_CORNERS_3D = None
        
        self.GEOM_DIM_X = None
        self.GEOM_DIM_Y = None
        self.GEOM_DIM_Z = None

        self.GEOM_SHAPE = None
        self.GEOM_CENTROID = None
        
        self.TOPO_REL_X = None
        self.TOPO_REL_Y = None
        self.TOPO_REL_Z = None

        self.TOPO_SUPERSTRUCTURE = None
        
        self.TOPO_ELS_ABOVE = []
        self.TOPO_ELS_BELOW = []
        
        self.BRIDGE_PART = None
        self.BRIDGE_SUBPART = None
    
    def __repr__(self):
        if self.GEOM_VERTICES is None:
            verts_label = "None"
        elif isinstance(self.GEOM_VERTICES, str) and self.GEOM_VERTICES == "N/A":
            verts_label = "N/A"
        else:
            verts_label = "Yes"

        return (
            f"IfcBridgeElement(GUID={self.ID_GUID}, "
            f"SEMA_CLASS:{self.SEMA_CLASS}, "
            f"SEMA_PTYPE:{self.SEMA_PTYPE}, "
            f"TOPO_SUPERSTRUCTURE:{self.TOPO_SUPERSTRUCTURE}"
            # f"GEOM_VERTS:{verts_label}, "
            # f"GEOM_CORNERS:{self.GEOM_CORNERS}"
            # f"GEOM_X:{self.GEOM_DIM_X}, GEOM_Y={self.GEOM_DIM_Y}, GEOM_Z={self.GEOM_DIM_Z}, "
            f"GEOM_SHAPE:{self.GEOM_SHAPE}, "
            # f"GEOM_CENTROID:{self.GEOM_CENTROID})"
            f"TOPO_REL_X:{self.TOPO_REL_X}"
            f"TOPO_REL_Y:{self.TOPO_REL_Y}"
            f"TOPO_REL_Z:{self.TOPO_REL_Z}"
            f"TOPO_ELS_ABOVE:{self.TOPO_ELS_ABOVE}"
            f"TOPO_ELS_BELOW:{self.TOPO_ELS_BELOW}"

        )

    def __getstate__(self):
        # Create a copy of the instance's __dict__
        state = self.__dict__.copy()
        # Remove or neutralize non-pickleable attributes
        state['IFC_ELEMENT'] = None
        state['SETTINGS'] = None
        state['SHAPE'] = None
        state['GEOMETRY'] = None
        return state

    def __setstate__(self, state):
        # Restore instance attributes
        self.__dict__.update(state)

    def get_representation(self):
        '''
        Checks if the ifc element has representation
        Return True or False
        '''
        if hasattr(self.IFC_ELEMENT,"Representation") and self.IFC_ELEMENT.Representation:
            if len(self.IFC_ELEMENT.Representation.Representations) > 0:
                return True
        else:
            return False
        
    def get_predefined_type(self):
        '''
        Gets the ifc element predefined type of an element if available
        Return an string representing the ifc predefined type
        '''
        if self.IFC_ELEMENT:
            try:
                predefined_type = ifcopenshell.util.element.get_predefined_type(self.IFC_ELEMENT)
                if predefined_type is not None:
                    self.SEMA_PTYPE = predefined_type
                else:
                    self.SEMA_PTYPE = None
            except Exception as e:
                raise ValueError(e)
        else:
            raise ValueError("No IfcElement found")
    
    def get_ifc_element_vertices(self):
        '''
        Gets the vertices of an ifc element if the shape is computable
        Return an numpy array with the elements vertices in global coordinates
        '''
        if self.IFC_ELEMENT and self.HAS_REPRESENTATION == True:
            try:
                shape = ifcopenshell.geom.create_shape(self.SETTINGS, self.IFC_ELEMENT)
                self.SHAPE = shape
                geometry = shape.geometry
                self.GEOMETRY = geometry
                element_vertices = ifcopenshell.util.shape.get_shape_vertices(shape, geometry)

                # Ensure that the vertex array is a numpy array of shape (N,3)
                element_vertices = np.asarray(element_vertices)
                if element_vertices.ndim != 2 or element_vertices.shape[1] != 3:
                    raise ValueError("Array of vertices must be of shape (N,3)")
                
                self.GEOM_VERTICES = element_vertices
            
            except Exception as e:
                print(e)
                self.GEOMETRY = "N/A"
                self.SHAPE = "N/A"
                self.GEOM_VERTICES = "N/A"
        else:
            self.GEOMETRY = "N/A"
            self.SHAPE = "N/A"
            self.GEOM_VERTICES = "N/A"
    
    def calculate_ifc_element_obb_z_vertical(self):
        '''
        Calculates the element object oriented box with the vertical Z
        Returns the element oriented bounding box corners in global coordinates,
        and the x_extent, y_extent and z_extent of the bounding box
        '''
        
        if isinstance(self.GEOM_VERTICES, str) and self.GEOM_VERTICES == "N/A":
            self.GEOM_DIM_X = "N/A"
            self.GEOM_DIM_Y = "N/A"
            self.GEOM_DIM_Z = "N/A"
            return
        
        else:
            points = np.asarray(self.GEOM_VERTICES)
            if points.size == 0:
                self.GEOM_DIM_X = "N/A"
                self.GEOM_DIM_Y = "N/A"
                self.GEOM_DIM_Z = "N/A"
                raise ValueError("The point array is empty")

            # 1) Separate the z-coordinate and get the min and max for z dimensions
            z_values = points[:, 2]
            z_min = np.min(z_values)
            z_max = np.max(z_values)
            dim_z = z_max - z_min

            # 2) Perform Principal Component Analysis (PCA) only for x and y
            points_xy = points[:, :2]
            # Compute the centroid in 2D
            mean_xy = np.mean(points_xy, axis=0)

            # Center the points around the centroid
            points_xy_centered = points_xy - mean_xy

            # Compute the 2x2 covariance matrix and its eigen-decomposition
            cov2x2 = np.cov(points_xy_centered, rowvar=0, bias=1)
            eigen_vals, eigen_vecs = np.linalg.eig(cov2x2)

            # Sort eigenvalues and the corresponding eigenvectors in descending order
            sort_idx = np.argsort(eigen_vals)[::-1]
            eigen_vals = eigen_vals[sort_idx]
            eigen_vecs = eigen_vecs[:, sort_idx]

            # 7) Rotate the centered points into principal axes space
            points_xy_rotated = points_xy_centered @ eigen_vecs

            # 8) Get min and max of the rotated points
            min_xy = np.min(points_xy_rotated, axis=0)
            max_xy = np.max(points_xy_rotated, axis=0)

            # 9) Dimensions in local (principal) coordinates
            dim_x = max_xy[0] - min_xy[0]
            dim_y = max_xy[1] - min_xy[1]

            # 10) Center of the bounding box in local (rotated) space
            center_xy_local = 0.5 * (min_xy + max_xy)

            # 11) The four corners of the bounding box in local (rotated) 2D
            x_min, y_min = min_xy
            x_max, y_max = max_xy

            corners_local_2d = np.array([
                [x_min, y_min],
                [x_min, y_max],
                [x_max, y_max],
                [x_max, y_min]
            ])

            # 12) Transform the 2D local corners back to the global XY
            corners_global_2d = (corners_local_2d @ eigen_vecs.T) + mean_xy

            # 13) Build the 3D bounding box by combining XY corners with z-min and z-max
            corners_world = []
            for corner2d in corners_global_2d:
                # Lower face (zmin)
                corners_world.append([corner2d[0], corner2d[1], z_min])
            for corner2d in corners_global_2d:
                # Upper face (zmax)
                corners_world.append([corner2d[0], corner2d[1], z_max])
            corners_world = np.array(corners_world)

            self.GEOM_CORNERS = corners_world
            self.GEOM_DIM_X = dim_x
            self.GEOM_DIM_Y = dim_y
            self.GEOM_DIM_Z = dim_z

    def calculate_ifc_element_obb_3d(self):
        """
        Calculates the fully 3D object-oriented bounding box of the element
        by performing PCA in three dimensions.

        Populates:
            self.GEOM_CORNERS: (8, 3) array of bounding box corners in global coords
            self.GEOM_DIM_X, self.GEOM_DIM_Y, self.GEOM_DIM_Z: the extents along principal axes
        """
        # 1) Handle cases with no geometry
        if isinstance(self.GEOM_VERTICES, str) and self.GEOM_VERTICES == "N/A":
            self.GEOM_DIM_X = "N/A"
            self.GEOM_DIM_Y = "N/A"
            self.GEOM_DIM_Z = "N/A"
            return

        points = np.asarray(self.GEOM_VERTICES)
        if points.size == 0:
            self.GEOM_DIM_X = "N/A"
            self.GEOM_DIM_Y = "N/A"
            self.GEOM_DIM_Z = "N/A"
            raise ValueError("The point array is empty")

        # 2) Compute centroid (mean) of the 3D points
        mean_3d = np.mean(points, axis=0)

        # 3) Center the points around the mean
        points_centered = points - mean_3d

        # 4) Compute the 3x3 covariance matrix
        #    rowvar=0 => each column is a variable (X, Y, Z)
        #    bias=1   => normalization by N instead of (N-1); either is fine for PCA shape analysis
        cov3x3 = np.cov(points_centered, rowvar=0, bias=1)

        # 5) Eigen-decompose the covariance matrix
        eigen_vals, eigen_vecs = np.linalg.eig(cov3x3)

        # 6) Sort eigenvalues (and corresponding eigenvectors) in descending order
        sort_idx = np.argsort(eigen_vals)[::-1]
        eigen_vals = eigen_vals[sort_idx]
        eigen_vecs = eigen_vecs[:, sort_idx]  # shape (3,3)

        # 7) Rotate the centered points into the principal axes space
        #    principal axes matrix: columns are eigen_vecs
        #    So for each point p_centered, p_rot = p_centered @ eigen_vecs
        #    => shape (N, 3)
        points_rotated = points_centered @ eigen_vecs

        # 8) Extract min and max extents along each principal axis
        min_xyz = np.min(points_rotated, axis=0)  # shape (3,)
        max_xyz = np.max(points_rotated, axis=0)  # shape (3,)

        # 9) Dimensions of the OBB in local (principal) space
        dim_x = max_xyz[0] - min_xyz[0]
        dim_y = max_xyz[1] - min_xyz[1]
        dim_z = max_xyz[2] - min_xyz[2]

        # 10) Build the 8 corners in local (rotated) space
        #     We'll form all combinations of (min_x, max_x), (min_y, max_y), (min_z, max_z)
        #     in a consistent order.
        x_min, y_min, z_min = min_xyz
        x_max, y_max, z_max = max_xyz

        corners_local = np.array([
            [x_min, y_min, z_min],
            [x_min, y_min, z_max],
            [x_min, y_max, z_min],
            [x_min, y_max, z_max],
            [x_max, y_min, z_min],
            [x_max, y_min, z_max],
            [x_max, y_max, z_min],
            [x_max, y_max, z_max]
        ])

        # 11) Transform these local corners back to the global coordinate system
        #     corners_global = (corners_local @ eigen_vecs.T) + mean_3d
        #     Because corners_local is shape (8,3), eigen_vecs.T is (3,3).
        corners_global = (corners_local @ eigen_vecs.T) + mean_3d

        # 12) Store results in the instance
        self.GEOM_CORNERS_3D = corners_global
        
    def evaluate_shape(self):
        """
        Evaluates the shape using the bounding box dimensions stored in self.
        If dimensions are None, triggers OBB calculation.
        """
        if (
            self.GEOM_DIM_X == "N/A" or 
            self.GEOM_DIM_Y == "N/A" or
            self.GEOM_DIM_Z == "N/A" or
            self.GEOM_DIM_X is None or
            self.GEOM_DIM_Y is None or
            self.GEOM_DIM_Z is None
        ):
            self.GEOM_SHAPE = "N/A"
            return self.GEOM_SHAPE
        
        try:
            x_val = float(self.GEOM_DIM_X)
            y_val = float(self.GEOM_DIM_Y)
            z_val = float(self.GEOM_DIM_Z)

        except (ValueError, TypeError):
            self.GEOM_SHAPE = "N/A"
            return self.GEOM_SHAPE

        dims = sorted([x_val, y_val, z_val])
        min_dim, mid_dim, max_dim = dims

        shape_label = "N/A"

        if max_dim / min_dim < 2:
            shape_label = "NEU_PRISM"
        # HOR_PLANAR: Horizontal Planar (slab-like element)
        elif (min_dim == z_val) and (x_val > 5 * z_val) and (y_val > 5 * z_val):
            shape_label = "HOR_PLANAR"
        # HOR_LINEAR: Horizontal Linear (beam or girder-like element)
        elif ((max_dim == x_val or max_dim == y_val) and (max_dim > 3 * mid_dim) and (mid_dim / min_dim < 4)):
            shape_label = "HOR_LINEAR"
        # VER_PLANAR (wall or Plate like element)
        elif (min_dim != z_val) and (min_dim * 5 < mid_dim) and (min_dim * 5 < max_dim):
            shape_label = "VER_PLANAR"
        # VER_LINEAR: Vertical Linear (column like element)
        elif (max_dim == z_val) and (z_val > 5 * y_val) and (max(x_val, y_val) / min(x_val, y_val) < 4):
            shape_label = "VER_LINEAR"
        self.GEOM_SHAPE = shape_label
        return self.GEOM_SHAPE
    
    def calculate_element_centroid(self):
        '''
        
        '''
        if self.SHAPE == "N/A":
            self.GEOM_CENTROID = "N/A"
        else:
            self.GEOM_CENTROID = ifcopenshell.util.shape.get_shape_bbox_centroid(self.SHAPE, self.GEOMETRY)

    def calculate_centroid_relative_position(self, model_obb_corners):
        '''
        Calculates the relative position of the element centroid in relation to the model obb
        Returns the following values:
            TOPO_X_RELATIVE between (-1,1)
            TOPO_Y_RELATIVE between (-1,1)
            TOPO_Z_RELATIVE between (0,1)
        '''
        if len(self.GEOM_CENTROID) == 0 or isinstance(self.GEOM_CENTROID,str):
            self.TOPO_REL_X = "N/A"
            self.TOPO_REL_Y = "N/A"
            self.TOPO_REL_Z ="N/A"
            
        elif isinstance(self.GEOM_CENTROID, np.ndarray):
            V = np.array (model_obb_corners, dtype= float)
            P = np.array (self.GEOM_CENTROID, dtype = float)

            if V.shape != (8,3):
                raise ValueError("Model OBB vertices must have shape (8,3)")
            center = np.mean(V, axis = 0)

            z_coords = V[:,2]
            min_z = np.min(z_coords)
            atolerance = 1e-06 #tolerance for floating point comparison
            bottom_face = V[np.isclose(z_coords,min_z,atol=atolerance)]

            if bottom_face.shape[0] != 4:
                raise ValueError(
                    f' Could not identify 4 bottom face vertices'
                    f' Found {bottom_face[0]} vertices'
                )
            b0 = bottom_face[0]
            b1 = bottom_face[1]
            u = (b1 - b0)[:2]  # Vector in XY plane
            u = u/ np.linalg.norm(u) #Normalize
            u = np.array([u[0], u[1], 0.0])  # Extend to 3D

            w = np.array([0.0, 0.0, 1.0])  # Global Z-axis
            v = np.cross(w, u)  # Perpendicular to both u and w

            # Projections of OBB vertices on each axis
            proj_u = np.dot(V - center, u)
            proj_v = np.dot(V - center, v)
            proj_w = np.dot(V - center, w)

            min_u, max_u = np.min(proj_u), np.max(proj_u)
            min_v, max_v = np.min(proj_v), np.max(proj_v)
            min_w, max_w = np.min(proj_w), np.max(proj_w)

            # Project centroid onto the OBB local axes
            P_local_u = np.dot((P - center), u)
            P_local_v = np.dot((P - center), v)
            P_local_w = np.dot((P - center), w)

            # Normalize the coordinates
            nx = 2 * ((P_local_u - min_u) / (max_u - min_u) - 0.5)  # Scale to [-1, 1]
            ny = 2 * ((P_local_v - min_v) / (max_v - min_v) - 0.5)  # Scale to [-1, 1]
            nz = (P_local_w - min_w) / (max_w - min_w)  # Scale to [0, 1]

            # Clamp the values to handle centroids outside the OBB
            nx = np.clip(nx, -1, 1)
            ny = np.clip(ny, -1, 1)
            nz = np.clip(nz, 0, 1)

            self.TOPO_REL_X = nx
            self.TOPO_REL_Y = ny
            self.TOPO_REL_Z = nz

            return np.array([nx, ny, nz])
        
    def calculate_centroid_relative_position_dynamic(self, model_obb_corners):
        '''
        Dynamically calculates the relative position of the element centroid in relation to the model OBB
        Returns the following values:
            TOPO_X_RELATIVE between (-1,1) relative to the longest side
            TOPO_Y_RELATIVE between (-1,1) relative to the shortest side
            TOPO_Z_RELATIVE between (0,1)
        '''
        if len(self.GEOM_CENTROID) == 0 or isinstance(self.GEOM_CENTROID, str):
            self.TOPO_REL_X = "N/A"
            self.TOPO_REL_Y = "N/A"
            self.TOPO_REL_Z = "N/A"
        elif isinstance(self.GEOM_CENTROID, np.ndarray):
            V = np.array(model_obb_corners, dtype=float)
            P = np.array(self.GEOM_CENTROID, dtype=float)
    
            if V.shape != (8, 3):
                raise ValueError("Model OBB vertices must have shape (8, 3)")
            center = np.mean(V, axis=0)
    
            z_coords = V[:, 2]
            min_z = np.min(z_coords)
            atolerance = 1e-06  # tolerance for floating-point comparison
            bottom_face = V[np.isclose(z_coords, min_z, atol=atolerance)]
    
            if bottom_face.shape[0] != 4:
                raise ValueError(
                    f'Could not identify 4 bottom face vertices. Found {bottom_face.shape[0]} vertices.'
                )
    
            # Compute edges and their lengths
            edges = [bottom_face[i] - bottom_face[(i + 1) % 4] for i in range(4)]
            edge_lengths = [np.linalg.norm(edge[:2]) for edge in edges]
    
            # Identify longest and shortest edges
            longest_edge_index = np.argmax(edge_lengths)
            shortest_edge_index = np.argmin(edge_lengths)
    
            # Compute axes u and v dynamically
            u = edges[longest_edge_index][:2]  # Longest edge
            u = u / np.linalg.norm(u)
            u = np.array([u[0], u[1], 0.0])
    
            v = edges[shortest_edge_index][:2]  # Shortest edge
            v = v / np.linalg.norm(v)
            v = np.array([v[0], v[1], 0.0])
    
            # Ensure axes form a right-handed system with the Z-axis
            w = np.array([0.0, 0.0, 1.0])  # Global Z-axis
            v = np.cross(w, u)
    
            # Projections of OBB vertices on each axis
            proj_u = np.dot(V - center, u)
            proj_v = np.dot(V - center, v)
            proj_w = np.dot(V - center, w)
    
            min_u, max_u = np.min(proj_u), np.max(proj_u)
            min_v, max_v = np.min(proj_v), np.max(proj_v)
            min_w, max_w = np.min(proj_w), np.max(proj_w)
    
            # Project centroid onto the OBB local axes
            P_local_u = np.dot((P - center), u)
            P_local_v = np.dot((P - center), v)
            P_local_w = np.dot((P - center), w)
    
            # Normalize the coordinates
            nx = 2 * ((P_local_u - min_u) / (max_u - min_u) - 0.5)  # Scale to [-1, 1]
            ny = 2 * ((P_local_v - min_v) / (max_v - min_v) - 0.5)  # Scale to [-1, 1]
            nz = (P_local_w - min_w) / (max_w - min_w)  # Scale to [0, 1]
    
            # Clamp the values to handle centroids outside the OBB
            nx = np.clip(nx, -1, 1)
            ny = np.clip(ny, -1, 1)
            nz = np.clip(nz, 0, 1)
    
            self.TOPO_REL_X = nx  # Relative to longest side
            self.TOPO_REL_Y = ny  # Relative to shortest side
            self.TOPO_REL_Z = nz
    
            return np.array([nx, ny, nz])
        
    # def calculate_centroid_relation_superstructure(self, plane_point, plane_normal, tolerance = 1e-6):
    #     '''
    #     '''
    #     # 1) If centroid is not available
    #     if (self.GEOM_CENTROID is None) or (
    #         isinstance(self.GEOM_CENTROID, str) and self.GEOM_CENTROID == "N/A"
    #     ):
    #         self.TOPO_SUPERSTRUCTURE = "UNKNOWN"
    #         return "UNKNOWN"
    
    #     # 2) Ensure the plane normal is oriented upward
    #     plane_point, plane_normal = self.orient_plane_normal_up(plane_point, plane_normal)
    
    #     # 3) Compute signed distance
    #     centroid_arr = np.asarray(self.GEOM_CENTROID, dtype=float)
    #     plane_point_arr = np.asarray(plane_point, dtype=float)
    
    #     # plane_normal is already unit length from orient_plane_normal_up
    #     d = np.dot(centroid_arr - plane_point_arr, plane_normal)
    
    #     # 4) Classify
    #     if d > tolerance:
    #         self.TOPO_SUPERSTRUCTURE = "ABOVE"
    #     elif d < -tolerance:
    #         self.TOPO_SUPERSTRUCTURE = "BELOW"
    #     else:
    #         self.TOPO_SUPERSTRUCTURE = "ON"
    
    #     return self.TOPO_SUPERSTRUCTURE

    def orient_plane_normal_up(self, plane_point, plane_normal):
        """
        Ensures that plane_normal points in the same general direction 
        as the global +Z axis (i.e., 'up').

        Returns:
            plane_point  (unchanged)
            plane_normal (possibly flipped if it was pointing down)
        """
        norm_len = np.linalg.norm(plane_normal)
        if norm_len < 1e-12:
            raise ValueError("Plane normal has zero length or is undefined.")
        plane_normal = plane_normal / norm_len

        # Dot product with global up
        global_up = np.array([0.0, 0.0, 1.0])
        if np.dot(plane_normal, global_up) < 0.0:
            # Flip the plane normal if it's pointing down
            plane_normal = -plane_normal

        return plane_point, plane_normal
    
    def classify_element_by_plane_intersection(self, plane_point, plane_normal, corners_3d, threshold=0.01):
        """
        Classifies an element as ABOVE, BELOW, or SUPERSTRUCTURE based on
        the bounding box corners (GEOM_CORNERS_3D) intersecting a plane.

        Args:
            plane_point  : (3,) A point on the plane.
            plane_normal : (3,) The normal to the plane.
            corners_3d   : (8,3) The OBB corners in global coordinates.
            threshold    : float, distance threshold to consider 'on' the plane.

        Returns:
            str: "ABOVE", "BELOW", or "SUPERSTRUCTURE"
        """
        # 1) Ensure plane_normal is oriented properly
        plane_point, plane_normal = self.orient_plane_normal_up(plane_point, plane_normal)

        # 2) Compute signed distances for each corner
        signed_dists = []
        for corner in corners_3d:
            d = np.dot(corner - plane_point, plane_normal)
            signed_dists.append(d)

        min_dist = min(signed_dists)
        max_dist = max(signed_dists)

        # 3) Classification logic
        if min_dist > 0:
            # Entire OBB is above the plane
            return "ABOVE"
        elif max_dist < -threshold:
            # Entire OBB is below the plane
            return "BELOW"
        else:
            # Some corners are above and some are below => intersects plane
            return "SUPERSTRUCTURE"

    def calculate_element_superstructure_relation_obbs(self, plane_point, plane_normal, threshold=0.01):
        """
        Uses the element's 3D OBB corners to classify if the element is
        ABOVE, BELOW, or SUPERSTRUCTURE with respect to the reference plane.

        Populates:
            self.TOPO_SUPERSTRUCTURE = "ABOVE", "BELOW", or "SUPERSTRUCTURE" 
        """
        # 1) Check if the OBB corners are available
        if not isinstance(self.GEOM_CORNERS_3D, np.ndarray) or self.GEOM_CORNERS_3D.shape != (8, 3):
            # If not available or invalid, mark UNKNOWN
            self.TOPO_SUPERSTRUCTURE = "UNKNOWN"
            return "UNKNOWN"

        # 2) Perform the plane intersection classification
        classification = self.classify_element_by_plane_intersection(
            plane_point, plane_normal, self.GEOM_CORNERS_3D, threshold=threshold
        )

        # 3) Store the result
        self.TOPO_SUPERSTRUCTURE = classification
        return classification
    
    def evaluate_bridge_part(self, bridge_length=None):
        for part_mapping in BRIDGE_PARTS:
            if self.SEMA_CLASS != part_mapping["class"]:
                continue  # Skip if the class does not match

            # Check conditions
            conditions = part_mapping.get("conditions", {})
            all_conditions_met = True

            for attr, condition in conditions.items():
                element_value = getattr(self, attr, None)

                if attr in ["TOPO_ELS_ABOVE", "TOPO_ELS_BELOW"] and isinstance(element_value, list):
                    # Extract IFC_CLASS from tuples in element_value
                    ifc_classes = [item[1] for item in element_value if isinstance(item, tuple)]
                    if None in condition and not ifc_classes:  # Empty list case
                        continue  # This condition is satisfied
                    # Check if any IFC_CLASS in the condition matches the extracted classes
                    if not any(cond in ifc_classes for cond in condition):
                        all_conditions_met = False
                        break
                elif attr =="TOPO_REL_X":
                    if isinstance(element_value,(int,float))and isinstance (condition, dict):
                        abs_value= abs(element_value)
                        if not (condition["min"]<= abs_value <= condition["max"]):
                            all_conditions_met = False
                            break
                elif isinstance(condition, list):
                    # Regular list comparison for non-tuple values
                    if element_value not in condition:
                        all_conditions_met = False
                        break
                elif callable(condition):
                    # Callable (e.g., for dynamic ranges like TOPO_REL_X)
                    range_values = condition(bridge_length)
                    if not (range_values["min"] <= element_value <= range_values["max"]):
                        all_conditions_met = False
                        break
                elif isinstance(condition, dict):
                    # Dictionary (e.g., TOPO_REL_X with min/max range)
                    if element_value is None or not (condition["min"] <= element_value <= condition["max"]):
                        all_conditions_met = False
                        break
                else:
                    # Single value equality check
                    if element_value != condition:
                        all_conditions_met = False
                        break

            if all_conditions_met:
                # Assign the bridge part and subpart
                self.BRIDGE_PART = part_mapping["bridge_part"]
                self.BRIDGE_SUBPART = part_mapping.get("bridge_subpart", None)
                return  # Exit once a match is found

        # Default case
        self.BRIDGE_PART = "UNKNOWN"
        self.BRIDGE_SUBPART = "UNKNOWN"