import os
import ifcopenshell
import multiprocessing
import time
from pprint import pprint

from ifc_bridge_element import IfcBridgeElement
from ifc_bridge_model import IfcBridgeModel

from ifc_visualize import *
from ifc_bridge_creation import traverse_ifc_bridge_aggregation
from ifc_bridge_aggregation import place_element_in_bridge_structure

def process_ifc_element(args):
    """
    Helper function to process a single IFC element.
    This function will be executed in a separate process.
    
    Args:
        args (tuple): A tuple containing the IFC element and settings.
        
    Returns:
        IfcBridgeElement: The processed IfcBridgeElement instance.
    """
    element, settings = args
    bridge_element = IfcBridgeElement(element, settings)
    bridge_element.get_ifc_element_vertices()
    bridge_element.calculate_ifc_element_obb_z_vertical()
    bridge_element.evaluate_shape()
    bridge_element.calculate_element_centroid()
    bridge_element.get_predefined_type()
    return bridge_element

# Global variables for worker processes
GLOBAL_IFC_MODEL = None
GLOBAL_SETTINGS = None

# Excluded classes
BRIDGE_ELEMENT_EXCLUDED_CLASSES = ["IfcSite","IfcGeographicElement","IfcEarthworksFill","IfcEarthworksElement","IfcEarthworksCut","IfcSurfaceFeature"]

def init_worker(ifc_path):
    """
    Initializer for each worker process.
    Opens the IFC model and sets up the geometry settings.
    
    Args:
        ifc_path (str): Path to the IFC file.
    """
    global GLOBAL_IFC_MODEL
    global GLOBAL_SETTINGS
    GLOBAL_IFC_MODEL = ifcopenshell.open(ifc_path)
    GLOBAL_SETTINGS = ifcopenshell.geom.settings()
    GLOBAL_SETTINGS.set(GLOBAL_SETTINGS.USE_WORLD_COORDS, True)

def process_ifc_element_worker(element_id):
    """
    Worker function to process a single IFC element by its GlobalId.
    
    Args:
        element_id (str): The GlobalId of the IFC element.
        
    Returns:
        IfcBridgeElement or None: The processed IfcBridgeElement instance, or None if excluded/error.
    """

    global GLOBAL_IFC_MODEL
    global GLOBAL_SETTINGS
    try:
        # Retrieve the element by GlobalId
        element = GLOBAL_IFC_MODEL.by_guid(element_id)
        if not element:
            print(f"Element with GlobalId {element_id} not found.")
            return None
        if element.is_a() in BRIDGE_ELEMENT_EXCLUDED_CLASSES:
            return None
        
        # Initialize the IfcBridgeElement
        bridge_element = IfcBridgeElement(element, GLOBAL_SETTINGS)
        bridge_element.get_predefined_type()
        bridge_element.get_ifc_element_vertices()
        bridge_element.calculate_ifc_element_obb_z_vertical()
        bridge_element.calculate_ifc_element_obb_3d()
        bridge_element.evaluate_shape()
        bridge_element.calculate_element_centroid()
        
        return bridge_element
    except Exception as e:
        print(f"Error processing element {element_id}: {e}")
        return None

if __name__ == "__main__":
    start_time = time.time()

    # Define your IFC file paths
    ifc_paths = [
        r"C:\00_TECNALIA\01_PROJECTS\TCRINI\00_CODE\RemapBridges\public\ifc_bridges\model01_girder.ifc",
        r"C:\00_TECNALIA\01_PROJECTS\TCRINI\00_CODE\RemapBridges\public\ifc_bridges\model02_tablero.ifc",
        r"C:\00_TECNALIA\01_PROJECTS\TCRINI\00_CODE\RemapBridges\public\ifc_bridges\model03_zubiot.ifc",
        r"C:\00_TECNALIA\01_PROJECTS\TCRINI\00_CODE\RemapBridges\public\ifc_bridges\model04_kit.ifc",
        r"C:\00_TECNALIA\01_PROJECTS\TCRINI\00_CODE\RemapBridges\public\ifc_bridges\model05_berriz-IFC4-IFC4X3.ifc",
        r"C:\00_TECNALIA\01_PROJECTS\TCRINI\00_CODE\RemapBridges\public\ifc_bridges\model06_PrecastAndCastConcrete.ifc",
        r"C:\00_TECNALIA\01_PROJECTS\TCRINI\00_CODE\RemapBridges\public\ifc_bridges\Viadotto Acerno.ifc"
    ]
    input_file = ifc_paths[1]

    output_path = r"C:\00_TECNALIA\01_PROJECTS\TCRINI\00_CODE\RemapBridges\public\ifc_bridges_test_class"
    base_name = os.path.basename(input_file)
    name = os.path.splitext(base_name)[0]

    new_output_path = os.path.join(output_path,f'{name}_class.ifc')
    
    # Open the IFC model to retrieve element IDs
    ifc_model_main = ifcopenshell.open(input_file)

    if ifc_model_main:
        ifc_elements = ifc_model_main.by_type("IfcElement")
        # Extract GlobalIds of elements excluding the specified classes
        element_ids = [
            element.GlobalId 
            for element in ifc_elements 
            if element.is_a() not in BRIDGE_ELEMENT_EXCLUDED_CLASSES
        ]
        print(len(element_ids))

        # Determine the number of processes to use
        max_processes = 10
        num_processes = min((multiprocessing.cpu_count())-1,max_processes)
        print(multiprocessing.cpu_count())

        # Create a multiprocessing Pool with initializer
        with multiprocessing.Pool(
            processes=num_processes, 
            initializer=init_worker, 
            initargs=(input_file,)
        ) as pool:
            # Map the processing function to the element IDs
            ifc_bridge_model_elements = pool.map(process_ifc_element_worker, element_ids)

        # Filter out None results (excluded elements or errors)
        ifc_bridge_model_elements = [
            element for element in ifc_bridge_model_elements if element is not None
        ]
        
        ifc_bridge_model = IfcBridgeModel(ifc_model_main,ifc_bridge_model_elements)
        ifc_bridge_obb_corners = ifc_bridge_model.calculate_model_obb_perfect()
        '''
        THE ORDER OF EXECUTION SHOULD BE IMPROVED
        '''
        for bridge_element  in ifc_bridge_model.ELEMENT_LIST:
            bridge_element.calculate_centroid_relative_position_dynamic(ifc_bridge_obb_corners)
        
        super = ifc_bridge_model.get_superstructure_elements()

        # for bridge_element in super:
        #     print ("SUPER",bridge_element)
        #     print("-"*20)

        points_3d = ifc_bridge_model.get_vertices_superstructure()
        super_obb, dim = ifc_bridge_model.calculate_superstructure_obb()
        # print("SUPER_OBB",super_obb)
        super_lowest = ifc_bridge_model.calculate_superstructure_lower_plane()

        centroid, normal = ifc_bridge_model.fit_plane_to_points()
        # interval_y,division = ifc_bridge_model.group_elements_by_x_relative()
        element_classification = ifc_bridge_model.classify_elements_by_x_relative()

        for bridge_element in ifc_bridge_model.ELEMENT_LIST:
            bridge_element.calculate_element_superstructure_relation_obbs(centroid,normal,threshold=2)

        ifc_bridge_model.compute_topology_above_below(z_tol=0.01)
        length = ifc_bridge_model.GEOM_OBB_DIM_X
        
        ifc_bridge = traverse_ifc_bridge_aggregation(ifc_model=ifc_model_main)
        for element in ifc_bridge_model.ELEMENT_LIST:
            element.evaluate_bridge_part(bridge_length=length)
            print(element.ID_GUID,element.SEMA_CLASS,element.BRIDGE_PART,element.BRIDGE_SUBPART)
            print(element.TOPO_SUPERSTRUCTURE,element.TOPO_REL_X,element.GEOM_SHAPE,element.SEMA_PTYPE)
            print(element.GEOM_DIM_X,element.GEOM_DIM_Y,element.GEOM_DIM_Z)
            print(f"ABOVE:{element.TOPO_ELS_ABOVE}",f"BELOW:{element.TOPO_ELS_BELOW}")
            print("-"*100)

            bridge_part_type = element.BRIDGE_PART if hasattr(element, 'BRIDGE_PART') else None
            bridge_subpart_type = element.BRIDGE_SUBPART if hasattr(element,"BRIDGE_SUBPART") else None
            element_guid = element.ID_GUID
            ifc_element = ifc_model_main.by_guid(element_guid)

            place_element_in_bridge_structure(
            ifc_model= ifc_model_main,
            ifc_bridge = ifc_bridge,
            element = ifc_element,
            bridge_part_type= bridge_part_type,
            bridge_subpart_type = bridge_subpart_type,
        )
        ifc_model_main.write(new_output_path)

        fig, ax = visualize_obb_and_plane(
            points_3d= points_3d,
            corners_obb= super_obb,
            corners_global_bb = ifc_bridge_obb_corners,
            plane_point=centroid,
            plane_normal= normal,
            plane_size=None,
            connect_obb_edges= True,
            label_global_bb= True
        )

        end_time = time.time()
        elapsed_time = end_time - start_time 
        print(f"Elapsed Time: {elapsed_time:.2f} seconds")
        plt.show()
    else:
        print("Failed to open IFC model.")