import numpy as np
import math
import ifcopenshell
import ifcopenshell.geom
import ifcopenshell.util.element
import ifcopenshell.util.shape

class IfcBridgeModel:
    def __init__(self,ifc_model,ifc_bridge_element_list):
        self.MODEL = ifc_model
        self.ELEMENT_LIST = ifc_bridge_element_list
        self.EXCLUDED_CLASSES = ["IfcBuildingElementProxy","IfcGeographicElement"]

        self.ELEMENT_CLASSIFICATION = None
        
        self.SUPERSTRUCTURE_LIST = None
        self.SUPERSTRUCTURE_CORNERS = None
        self.SUPERSTRUCTURE_OBB_CORNERS = None
        self.SUPERSTRUCTURE_OBB_DIMENSIONS = None

        self.SUPERSTRUCTURE_OBB_LOWEST = None
        
        self.GEOM_MODEL_VERTICES = self.get_vertices_except_classes()
        self.GEOM_OBB_CORNERS = None
        self.GEOM_OBB_CENTER = None

        self.GEOM_OBB_DIM_X = None
        self.GEOM_OBB_DIM_Y = None
        self.GEOM_OBB_DIM_Z = None

    def __repr__(self) -> str:
        return (
            f"IfcBridgeModel("
            f"MODEL={self.MODEL}, "
            f"ELEMENT_LIST_COUNT={len(self.ELEMENT_LIST)}, "
            f"GEOM_OBB_CORNERS={self.GEOM_OBB_CORNERS if self.GEOM_OBB_CORNERS is not None else 'Not Calculated'}, "
            f"GEOM_OBB_CENTER={self.GEOM_OBB_CENTER if self.GEOM_OBB_CENTER is not None else 'Not Calculated'}, "
            f"GEOM_OBB_DIM_X={self.GEOM_OBB_DIM_X if self.GEOM_OBB_DIM_X is not None else 'Not Calculated'}, "
            f"GEOM_OBB_DIM_Y={self.GEOM_OBB_DIM_Y if self.GEOM_OBB_DIM_Y is not None else 'Not Calculated'}, "
            f"GEOM_OBB_DIM_Z={self.GEOM_OBB_DIM_Z if self.GEOM_OBB_DIM_Z is not None else 'Not Calculated'})"
        )

    def get_vertices_except_classes(self)-> np.ndarray:
        '''
        Get the corners of elements inside the list except the excluded classes
        Return the flat list of all corners
        '''
        if not self.ELEMENT_LIST:
            raise ValueError("Element list is empty!")
        
        model_corners = []
        if len(self.ELEMENT_LIST)!= 0 :
            for ifc_bridge_element in self.ELEMENT_LIST:
                if (
                isinstance(ifc_bridge_element.GEOM_CORNERS, np.ndarray) 
                and len(ifc_bridge_element.GEOM_CORNERS) != 0 
                and ifc_bridge_element.SEMA_CLASS not in self.EXCLUDED_CLASSES
                ):
                    element_corners = ifc_bridge_element.GEOM_CORNERS
                    model_corners.append(element_corners)
        if model_corners:
            return np.vstack(model_corners)
        else:
            return np.empty((0,))
    

    def calculate_model_obb_perfect(self):
        '''
        Calculates the model oriented bounding box based on the corners of the input elements
        Ensures that the obb has horizontal and vertical orthogonal faces
        Returns the model obb: cornerns, center, x_entent, y_extent and z_extent
        '''
        points = np.asarray(self.GEOM_MODEL_VERTICES)

        if points.size == 0:
            self.GEOM_OBB_CORNERS = None
            self.GEOM_OBB_CENTER = None
            self.GEOM_OBB_DIM_X = None
            self.GEOM_OBB_DIM_Y = None
            self.GEOM_OBB_DIM_Z = None
            return
        
        # Calculate the mean of the points
        means = np.mean(points, axis=0)

        # Calculate the covariance matrix for the XY plane
        cov = np.cov(points[:, :2], rowvar=0, bias=1)  # Only consider the XY plane for orientation

        # Eigen decomposition of the covariance matrix
        eigenvalues, eigenvectors = np.linalg.eig(cov)

        # Sort eigenvectors by eigenvalues to ensure major axes are principal directions
        sorted_indices = np.argsort(-eigenvalues)
        primary_axis = eigenvectors[:, sorted_indices[0]]  # Principal direction 1 (longest)
        secondary_axis = eigenvectors[:, sorted_indices[1]]  # Principal direction 2 (orthogonal)

        # Construct a right-handed coordinate system with the global Z-axis
        u = np.append(primary_axis, 0.0)  # Extend to 3D (XY plane direction)
        v = np.append(secondary_axis, 0.0)  # Extend to 3D (XY plane direction)
        w = np.array([0.0, 0.0, 1.0])  # Global Z-axis

        # Construct the transformation matrix
        transform = np.column_stack((u, v, w))

        # Rotate points to align with principal axes in the XY plane and Z-axis globally
        points_r = np.dot(points - means, transform)

        # Get the min and max coordinates along each axis
        co_min = np.min(points_r, axis=0)
        co_max = np.max(points_r, axis=0)

        # Define dimensions along each axis
        x_extent = co_max[0] - co_min[0]
        y_extent = co_max[1] - co_min[1]
        z_extent = co_max[2] - co_min[2]

        # Calculate the center in local coordinates
        center_local = (co_max + co_min) / 2

        # Define the corners of the bounding box in local space
        corners_local = np.array([
            [co_min[0], co_min[1], co_min[2]],  # Bottom face
            [co_min[0], co_max[1], co_min[2]],
            [co_max[0], co_max[1], co_min[2]],
            [co_max[0], co_min[1], co_min[2]],
            [co_min[0], co_min[1], co_max[2]],  # Top face
            [co_min[0], co_max[1], co_max[2]],
            [co_max[0], co_max[1], co_max[2]],
            [co_max[0], co_min[1], co_max[2]],
        ])

        # Transform corners back to world space
        corners_world = np.dot(corners_local, transform.T) + means

        # Transform the center back to world space
        center_world = np.dot(center_local, transform.T) + means

        # Return corners, center, and dimensions
        self.GEOM_OBB_DIM_X = x_extent
        self.GEOM_OBB_DIM_Y = y_extent
        self.GEOM_OBB_DIM_Z = z_extent

        self.GEOM_OBB_CORNERS = corners_world
        self.GEOM_OBB_CENTER = center_world
        
        return corners_world
    '''
    GETTING THE SUPERSTRUCTURE OF THE BRIDGE
    '''
    def get_superstructure_elements(self):
        '''
        This needs to be refined
        '''
        
        if self.GEOM_OBB_DIM_X < 50:
            bridge_rel_x_threshold = 0.5
        elif self.GEOM_OBB_DIM_X < 100:
            bridge_rel_x_threshold = 0.6
        elif self.GEOM_OBB_DIM_X < 200:
            bridge_rel_x_threshold = 0.8
        else:
            bridge_rel_x_threshold = 0.9


        superstructure_excluded_classes = ["IfcBuildingElementProxy","IfcSite","IfcWall","IfcWallStandardCase",
                                           "IfcPlate","IfcBeam","IfcFooting","IfcBearing","IfcMember","IfcVibrationDamper",
                                           "IfcVibrationIsolator","IfcRailing","IfcSurfaceFeature","IfcAnnotation","IfcMechanicalFastener"]
        bridge_superstructure_list = []
        if len(self.ELEMENT_LIST)!= 0:
            for bridge_element in self.ELEMENT_LIST:
                if isinstance(bridge_element.TOPO_REL_Z,str):
                    continue

                is_hor_planar = (
                    bridge_element.GEOM_SHAPE =="HOR_PLANAR" and
                    bridge_element.SEMA_CLASS not in superstructure_excluded_classes and
                    abs(bridge_element.TOPO_REL_X)<bridge_rel_x_threshold
                )
                is_slab = (
                    bridge_element.SEMA_CLASS == "IfcSlab" and
                    (bridge_element.GEOM_SHAPE != "VER_LINEAR" or bridge_element.GEOM_SHAPE != "VER_PLANAR") and
                    abs(bridge_element.TOPO_REL_X)<bridge_rel_x_threshold)

                if (is_hor_planar or is_slab) and bridge_element.TOPO_REL_Z> 0.5:
                    bridge_superstructure_list.append(bridge_element)
            
            self.SUPERSTRUCTURE_LIST = bridge_superstructure_list
        
        return bridge_superstructure_list
    
    def get_vertices_superstructure(self) ->np.ndarray:
        '''
        Get the corners of elements inside the superstructure list
        Return the flat list of all corners
        '''
        if not self.SUPERSTRUCTURE_LIST:
            raise ValueError("Superstructure list is empty")
        superstructure_vertices = []
        if len(self.SUPERSTRUCTURE_LIST)!=0:
            for ifc_bridge_element in self.SUPERSTRUCTURE_LIST:
                if(
                isinstance(ifc_bridge_element.GEOM_VERTICES, np.ndarray)
                and len(ifc_bridge_element.GEOM_VERTICES)!= 0 
                ):
                    element_vertices = ifc_bridge_element.GEOM_VERTICES
                    superstructure_vertices.append(element_vertices)

        if superstructure_vertices:
            self.SUPERSTRUCTURE_CORNERS = np.vstack(superstructure_vertices)
            return self.SUPERSTRUCTURE_CORNERS
        else:
            return np.empty((0,))
    
    def calculate_superstructure_obb(self):
        """
        Computes a 3D Oriented Bounding Box (OBB) for the given point cloud (N x 3),
        using PCA. Each local axis (x', y', z') is one of the principal directions.

        Returns:
            corners_world: (8, 3) array of the OBB corner coordinates in global space.
            center_world:  (3,)   the OBB center in global coordinates.
            dimensions:    (dx', dy', dz') the box lengths along each local axis.
            rotation_mat:  (3, 3) each column is one local axis in global coords.
                           - column 0 = x'
                           - column 1 = y'
                           - column 2 = z'
        """

        # Handle empty input
        points = self.SUPERSTRUCTURE_CORNERS
        # if points.shape[0] == 0:
        #     return None, None, None, None

        # 1) Compute centroid
        centroid = np.mean(points, axis=0)

        # 2) Center the points
        centered = points - centroid  # shape (N,3)

        # 3) 3x3 covariance matrix of the data
        cov = np.cov(centered, rowvar=False, bias=True)

        # 4) Eigen-decomposition
        eigvals, eigvecs = np.linalg.eig(cov)

        # 5) Sort by descending eigenvalue: largest -> smallest
        order = np.argsort(eigvals)[::-1]
        eigvals = eigvals[order]
        eigvecs = eigvecs[:, order]

        # -----------------------------------------------------------------------
        # At this point, 'eigvecs' has shape (3,3).
        # columns: eigvecs[:,0], eigvecs[:,1], eigvecs[:,2]
        # The first column is the direction of highest variance  (local x'),
        # the second column is the direction of second variance (local y'),
        # the third column is the direction of smallest variance (local z').
        #
        # If you want to label them differently—e.g. force local z' to be the
        # largest variance direction—just reorder columns here as you prefer.
        # -----------------------------------------------------------------------

        # 6) Define R = eigvecs as the rotation from local->world.
        #    So local_coords = R^T * (p - centroid).
        R = eigvecs  # shape (3,3)

        # 7) Transform all points into the local coordinate system
        #    local_coords[i] = R^T @ (points[i] - centroid)
        #    (N,3) = (N,3) @ (3,3)
        local_coords = (points - centroid) @ R

        # 8) Find min & max along each local axis
        min_local = np.min(local_coords, axis=0)
        max_local = np.max(local_coords, axis=0)

        # 9) Dimensions along each local axis
        dx = max_local[0] - min_local[0]
        dy = max_local[1] - min_local[1]
        dz = max_local[2] - min_local[2]
        dimensions = (dx, dy, dz)

        # 10) Build 8 corners in local space (combinations of min/max)
        corners_local = np.array([
            [min_local[0], min_local[1], min_local[2]],
            [min_local[0], min_local[1], max_local[2]],
            [min_local[0], max_local[1], min_local[2]],
            [min_local[0], max_local[1], max_local[2]],
            [max_local[0], min_local[1], min_local[2]],
            [max_local[0], min_local[1], max_local[2]],
            [max_local[0], max_local[1], min_local[2]],
            [max_local[0], max_local[1], max_local[2]],
        ])

        # 11) Local-space center
        center_local = 0.5 * (min_local + max_local)

        # 12) Transform corners and center back to global space:
        #
        #     corners_world[i] = centroid + R @ corners_local[i]
        #
        # Note: local_coords = R^T (points - centroid),
        #       so the inverse transform is points = centroid + R (local_coords).
        corners_world = corners_local @ R.T + centroid
        # center_world = center_local @ R.T + centroid
        
        self.SUPERSTRUCTURE_OBB_CORNERS = corners_world
        self.SUPERSTRUCTURE_OBB_DIMENSIONS = dimensions

        return corners_world, dimensions
    
    def calculate_superstructure_lower_plane(self):
        if self.SUPERSTRUCTURE_OBB_CORNERS.shape !=(8,3):
            raise ValueError("Input corners array must have shape(8,3)")
        
        xy_divisions = {}
        for i ,(x,y,z) in enumerate (self.SUPERSTRUCTURE_OBB_CORNERS):
            rounded_x = round(x,1)
            rounded_y = round(y,1)
            group_key = (rounded_x,rounded_y)

            if group_key not in xy_divisions:
                xy_divisions[group_key] = []
            xy_divisions[group_key].append(i)
        
        bottom_indices = []
        for group_key, idx_list in xy_divisions.items():
            # If the box is a proper OBB, each group should have 2 corners:
            # one for bottom, one for top. We pick the bottom by min z.
            # However, if there's floating mismatch or unusual geometry,
            # you might have more or fewer than 2.
            min_z_index = min(idx_list, key=lambda idx: self.SUPERSTRUCTURE_OBB_CORNERS[idx, 2])
            bottom_indices.append(min_z_index)

        # 4) Ensure we have exactly 4 unique bottom corners
        bottom_indices = list(set(bottom_indices))  # remove duplicates if any
        
        # if len(bottom_indices) != 4:
        #     raise ValueError(
        #         f"Did not find exactly 4 bottom corners. Found {len(bottom_indices)}. "
        #         "Check if the input corners really form a valid OBB, or increase rounding tolerance."
        #     )

        # 5) Gather the bottom corners
        bottom_face = self.SUPERSTRUCTURE_OBB_CORNERS[bottom_indices]
        self.SUPERSTRUCTURE_OBB_LOWEST = bottom_face
        return bottom_face
    
    def fit_plane_to_points(self):
        """
        Returns (plane_point, plane_normal) by least-squares plane fitting.
        plane_point is a point on the plane (the centroid).
        plane_normal is a unit vector normal to the plane.
        """

        points3d = self.SUPERSTRUCTURE_OBB_LOWEST
        # 1) Compute centroid
        centroid = np.mean(points3d, axis=0)
        # 2) Build matrix of zero-mean coords
        M = points3d - centroid  # shape (N,3)
        # 3) SVD or eigen on the covariance
        #    The plane normal is the smallest singular vector
        U, S, Vt = np.linalg.svd(M, full_matrices=False)
        # Right singular vectors are in Vt; the smallest singular value is S[-1]
        # The corresponding vector is Vt[-1], but we want the normal (which is last row of Vt).
        normal = Vt[-1, :]

        # Make sure it's unit length
        normal = normal / np.linalg.norm(normal)

        return centroid, normal
    
    '''
    DIVIDE THE MODEL IN ORDER TO REDUCE THE ALGORITH COMPLEXITY OF COMPUTING THE ABOVE AND BELOW RELATIONS
    '''
    def group_elements_by_x_relative(self, overlapping = 0.2):
        """
        Group elements based on the X component (TOPO_REL_X) of their relative position.
        For IfcColumn elements, divide the range [-1, 1] using the divisions (splitting points).

        Args:
            tolerance: Float, the maximum distance between points to treat them as the same group.

        Returns:
            ranges: A list of tuples representing ranges, where each range is defined by scaled divisions within [-1, 1].
            x_positions: The original X positions after rounding and grouping.
        """
        tolerance = 0.1
        if len(self.ELEMENT_LIST) == 0:
            return []  # Return empty list if no elements exist

        # Check if there are any IfcColumn elements
        if any(el.SEMA_CLASS == "IfcColumn" for el in self.ELEMENT_LIST):
            # Perform grouping for IfcColumn elements
            x_positions = []
            for el in self.ELEMENT_LIST:
                if el.SEMA_CLASS == "IfcColumn" and isinstance(el.TOPO_REL_X, (float, int)):
                    x_positions.append(round(el.TOPO_REL_X, 2))  # Round to reduce floating-point errors

            if len(x_positions) == 0:
                return []  # Return empty list if no valid X positions are found

            # Sort the X positions
            x_positions = sorted(set(x_positions))  # Remove duplicates and sort
            print("Sorted X positions:", x_positions)

            # Merge positions within the tolerance
            divisions = []
            for x in x_positions:
                # If divisions is empty or the difference between the current position and the last group
                # is greater than the tolerance, create a new group
                if not divisions or abs(x - divisions[-1]) > tolerance:
                    divisions.append(x)
                else:
                    # Merge into the existing group by averaging
                    divisions[-1] = (divisions[-1] + x) / 2  # Update the last group with the average
            print("Merged divisions:", divisions)

            # Ensure -1 and 1 are included as boundaries
            divisions = [-1.0] + divisions + [1.0]
            print("Final divisions with boundaries:", divisions)

            # Create ranges using the divisions
            ranges = [(float(divisions[i]), float(divisions[i + 1])) for i in range(len(divisions) - 1)]

            return ranges

        else:
            if isinstance(self.GEOM_OBB_DIM_X, float):
                # Calculate the number of intervals
                interval_number = int(round(self.GEOM_OBB_DIM_X / 20))

                # Ensure interval_number is at least 1 to avoid division by zero
                if interval_number < 1:
                    interval_number = 1

                # Calculate the step size for each interval
                step = 2.0 / interval_number  # Total range is 2 from -1 to 1

                # Generate the divisions
                divisions = [-1.0 + i * step for i in range(interval_number + 1)]

                # Create ranges using the divisions
                ranges = [(divisions[i], divisions[i + 1]) for i in range(len(divisions) - 1)]

                return ranges
            
    def group_elements_by_x_relative_overlap(self):
        """
        Group elements based on the X component (TOPO_REL_X) of their relative position.
        For IfcColumn elements, divide the range [-1, 1] using the divisions (splitting points) with overlapping.

        Args:
            overlapping: Float, the amount of overlap to add to each range.

        Returns:
            ranges: A list of tuples representing ranges, where each range is defined by scaled divisions within [-1, 1].
            x_positions: The original X positions after rounding and grouping.
        """
        overlapping = 0.1
        tolerance = 0.2
        if len(self.ELEMENT_LIST) == 0:
            return []  # Return empty list if no elements exist

        # Check if there are any IfcColumn elements
        if any(el.SEMA_CLASS == "IfcColumn" for el in self.ELEMENT_LIST):
            # Perform grouping for IfcColumn elements
            x_positions = []
            for el in self.ELEMENT_LIST:
                if el.SEMA_CLASS == "IfcColumn" and isinstance(el.TOPO_REL_X, (float, int)):
                    x_positions.append(round(el.TOPO_REL_X, 2))  # Round to reduce floating-point errors

            if len(x_positions) == 0:
                return []  # Return empty list if no valid X positions are found

            # Sort the X positions
            x_positions = sorted(set(x_positions))  # Remove duplicates and sort
            print("Sorted X positions:", x_positions)

            # Merge positions within the tolerance
            divisions = []
            for x in x_positions:
                # If divisions is empty or the difference between the current position and the last group
                # is greater than the tolerance, create a new group
                if not divisions or abs(x - divisions[-1]) > tolerance:
                    divisions.append(x)
                else:
                    # Merge into the existing group by averaging
                    divisions[-1] = (divisions[-1] + x) / 2  # Update the last group with the average
            print("Merged divisions:", divisions)

            # Ensure -1 and 1 are included as boundaries
            divisions = [-1.0] + divisions + [1.0]
            print("Final divisions with boundaries:", divisions)

            # Create ranges with overlapping
            ranges = []
            for i in range(len(divisions) - 1):
                start = divisions[i] - overlapping if i > 0 else divisions[i]  # No overlap for the first range start
                end = divisions[i + 1] + overlapping if i < len(divisions) - 2 else divisions[i + 1]  # No overlap for the last range end
                ranges.append((float(start), float(end)))

            return ranges

        else:
            if isinstance(self.GEOM_OBB_DIM_X, float):
                # Calculate the number of intervals
                interval_number = int(round(self.GEOM_OBB_DIM_X / 15))

                # Ensure interval_number is at least 1 to avoid division by zero
                if interval_number < 1:
                    interval_number = 1

                # Calculate the step size for each interval
                step = 2.0 / interval_number  # Total range is 2 from -1 to 1

                # Generate the divisions
                divisions = [-1.0 + i * step for i in range(interval_number + 1)]

                # Create ranges with overlapping
                ranges = []
                for i in range(len(divisions) - 1):
                    start = divisions[i] - overlapping if i > 0 else divisions[i]  # No overlap for the first range start
                    end = divisions[i + 1] + overlapping if i < len(divisions) - 2 else divisions[i + 1]  # No overlap for the last range end
                    ranges.append((float(start), float(end)))

                return ranges 

    def classify_elements_by_x_relative(self):
        '''

        '''
        ranges = self.group_elements_by_x_relative_overlap()
        if not ranges:
            self.ELEMENT_CLASSIFICATION = {}
            return {}
        
        classification = {r:[] for r in ranges}
        for element in self.ELEMENT_LIST:
            if hasattr(element, "TOPO_REL_X") and isinstance (element.TOPO_REL_X, (float, int)):
                #Check which range the element belongs to:
                for r in ranges:
                    if r[0] <= element.TOPO_REL_X <r[1]:
                        classification[r].append(element)
        
        self.ELEMENT_CLASSIFICATION = classification
        return classification
    '''
    COMPUTE ABOVE AND BELOW TOPOLOGIES PER EACH ELEMENT
    '''
    def check_xy_and_z_overlap(self, elem1, elem2, z_tol=0.05):
        """
        1) Check true overlap in XY plane (using polygons derived from OBB corners + SAT).
        2) Check if their Z bounding boxes intersect or come within z_tol.

        Returns:
            "above" if elem1 should be classified as above elem2
            "below" if elem1 should be classified as below elem2
            None if no overlap or if there's a gap > z_tol in the Z direction
        """
        # ---------------------------
        # 1) Check XY overlap
        # ---------------------------
        poly1 = [(c[0], c[1]) for c in elem1.GEOM_CORNERS_3D]
        poly2 = [(c[0], c[1]) for c in elem2.GEOM_CORNERS_3D]
        if not self.polygons_intersect_2d(poly1, poly2):
            return None  # No plan (XY) intersection => no relationship

        # ---------------------------
        # 2) Check Z bounding boxes
        # ---------------------------
        zs1 = [c[2] for c in elem1.GEOM_CORNERS_3D]
        zs2 = [c[2] for c in elem2.GEOM_CORNERS_3D]
        min_z1, max_z1 = min(zs1), max(zs1)
        min_z2, max_z2 = min(zs2), max(zs2)

        # If there's a gap > z_tol, no relationship
        # (elem1 is far below elem2, or elem2 is far below elem1)
        if max_z1 < (min_z2 - z_tol) or max_z2 < (min_z1 - z_tol):
            return None

        # At this point, they overlap in Z or are within z_tol. Decide "above" vs. "below"
        # A simple criterion: whichever has the higher bottom (min_z) is "above".
        if min_z1 >= min_z2:
            return "above"
        else:
            return "below"

    def polygons_intersect_2d(self, cornersA, cornersB):
        """
        Check intersection of two polygons in 2D using the Separating Axis Theorem (SAT).
        """
        def get_axes(corners):
            """Generate unique perpendicular axes from polygon edges."""
            axes = []
            n = len(corners)
            for i in range(n):
                x1, y1 = corners[i]
                x2, y2 = corners[(i + 1) % n]
                edge = (x2 - x1, y2 - y1)
                perp_axis = (edge[1], -edge[0])  # Perp to edge
                length = math.hypot(perp_axis[0], perp_axis[1])
                if length > 1e-12:
                    perp_axis = (perp_axis[0] / length, perp_axis[1] / length)
                axes.append(perp_axis)
            return axes

        def project_corners(corners, axis):
            """Project corners onto a given axis."""
            dots = []
            for cx, cy in corners:
                dots.append(cx * axis[0] + cy * axis[1])
            return min(dots), max(dots)

        # Combine axes from both polygons
        axes = get_axes(cornersA) + get_axes(cornersB)

        for axis in axes:
            minA, maxA = project_corners(cornersA, axis)
            minB, maxB = project_corners(cornersB, axis)
            if maxA < minB or maxB < minA:  # Gap found -> no intersection
                return False
        return True

    def compute_topology_above_below(self, z_tol=0.2):
        """
        For each classification group, find which elements:
          - Overlap in XY, and
          - Are within z_tol (or overlapping) in Z.

        Sets:
          elem.TOPO_ELS_ABOVE = [list of tuples (GUID, SEMA_CLASS)]
          elem.TOPO_ELS_BELOW = [list of tuples (GUID, SEMA_CLASS)]
        """
        if not self.ELEMENT_CLASSIFICATION:
            print("ELEMENT_CLASSIFICATION is empty. Please classify elements first.")
            return

        # Iterate over classification groups
        for range_key, elements in self.ELEMENT_CLASSIFICATION.items():
            n = len(elements)

            # Reset relationships in this group
            for elem in elements:
                elem.TOPO_ELS_ABOVE = []
                elem.TOPO_ELS_BELOW = []

            # Compare elements pairwise
            for i in range(n):
                e_i = elements[i]
                if e_i.GEOM_CORNERS_3D is None or len(e_i.GEOM_CORNERS_3D) == 0:
                    continue

                for j in range(i + 1, n):
                    e_j = elements[j]
                    if e_j.GEOM_CORNERS_3D is None or len(e_j.GEOM_CORNERS_3D) == 0:
                        continue

                    # Check combined XY overlap and refined Z logic
                    relation = self.check_xy_and_z_overlap(e_i, e_j, z_tol=z_tol)
                    if relation == "above":
                        # e_i above e_j
                        e_i.TOPO_ELS_BELOW.append((e_j.ID_GUID, e_j.SEMA_CLASS))
                        e_j.TOPO_ELS_ABOVE.append((e_i.ID_GUID, e_i.SEMA_CLASS))
                    elif relation == "below":
                        # e_i below e_j
                        e_j.TOPO_ELS_BELOW.append((e_i.ID_GUID, e_i.SEMA_CLASS))
                        e_i.TOPO_ELS_ABOVE.append((e_j.ID_GUID, e_j.SEMA_CLASS))

        # Finally, consolidate results into self.ELEMENT_LIST
        self.remap_classification_element_list()
        
    def remap_classification_element_list(self):
        '''
        Remap elements from the ELEMENT_CLASSIFICATION inside ELEMENT_LIST for the above and below relations
        '''
        element_map = {}
        for elem in self.ELEMENT_LIST:
            element_map[elem.ID_GUID] = elem
        
        for range_key in self.ELEMENT_CLASSIFICATION:
            elements = self.ELEMENT_CLASSIFICATION[range_key]

            for classified_elem in elements:
                if classified_elem.ID_GUID not in element_map:
                    continue
                actual_element = element_map[classified_elem.ID_GUID]

                for above in classified_elem.TOPO_ELS_ABOVE:
                    if above not in actual_element.TOPO_ELS_ABOVE:
                        actual_element.TOPO_ELS_ABOVE.append(above)
                
                for below in classified_elem.TOPO_ELS_BELOW:
                    if below not in actual_element.TOPO_ELS_BELOW:
                        actual_element.TOPO_ELS_BELOW.append(below)
                