import ifcopenshell
import ifcopenshell.guid
import os

def get_ifc_project(ifc_model):
    projects = ifc_model.by_type("IfcProject")
    if not projects:
        raise ValueError("No IfcProject found in the IFC Model")
    if len(projects)>1:
        raise ValueError("More than one IfcProjects found ")

def create_bridge_under_project(ifc_model, bridge_name="New Bridge", bridge_description="Bridge Entity"):
    """
    Ensures the IFC model contains only an IfcProject and adds an IfcBridge under it.

    Args:
        ifc_file: The IFC model instance loaded with ifcopenshell.
        bridge_name: The name to assign to the new IfcBridge.
        bridge_description: A description for the new IfcBridge.

    Returns:
        The created IfcBridge entity.
    """
    # Step 1: Retrieve all IfcProject entities
    projects = ifc_model.by_type("IfcProject")
    if not projects:
        raise ValueError("No IfcProject found in the IFC model.")
    project = projects[0]  # Assuming the first IfcProject

    # Step 2: Remove all existing spatial structures except IfcProject
    # This includes IfcSite, IfcBuilding, etc.
    
    # spatial_containers = ["IfcSite", "IfcBuilding", "IfcBuildingStorey", "IfcSpace"]
    # for container_type in spatial_containers:
    #     for container in ifc_model.by_type(container_type):
    #         # Remove all aggregate relationships
    #         rels = ifc_model.get_inverse(container)
    #         for rel in rels:
    #             if rel.is_a("IfcRelAggregates") and rel.RelatingObject == container:
    #                 ifc_model.remove(rel)
    #         # Finally, remove the container itself
    #         ifc_model.remove(container)

    # Step 3: Create IfcBridge entity
    # Ensure OwnerHistory exists

    owner_histories = ifc_model.by_type("IfcOwnerHistory")
    if not owner_histories:
        # Create a default OwnerHistory if none exists
        owner_history = ifc_model.create_entity("IfcOwnerHistory")
        ifc_model.add(owner_history)
    else:
        owner_history = owner_histories[0]

    bridge = ifc_model.create_entity(
        "IfcBridge",
        GlobalId=ifcopenshell.guid.new(),
        OwnerHistory=owner_history,
        Name=bridge_name,
        Description=bridge_description,
        ObjectType="Bridge",
        LongName=bridge_name
    )
    ifc_model.add(bridge)

    # Step 4: Relate the IfcBridge to the IfcProject using IfcRelAggregates
    rel_aggregates = ifc_model.create_entity(
        "IfcRelAggregates",
        GlobalId=ifcopenshell.guid.new(),
        OwnerHistory=owner_history,
        RelatingObject=project,
        RelatedObjects=[bridge]
    )
    ifc_model.add(rel_aggregates)

    return bridge

def assign_new_container(ifc_model, element_guid, new_container_class, bridge_part_name, bridge_part_type_enum):
    """
    Reassigns an element to a new container class (e.g., IfcBridgePart) while preserving relationships.
    If a container with the specified name already exists, reuse it.

    Args:
        ifc_model: The IFC model instance.
        element_guid: GUID of the element to reassign.
        new_container_class: Target class name for the new container (e.g., "IfcBridgePart").
        bridge_part_name: Name for the new container.
        bridge_part_type_enum: Type enum for the new container.

    Returns:
        The container entity (existing or newly created).
    """
    # Retrieve the element by GUID
    element = ifc_model.by_guid(element_guid)
    if not element:
        raise ValueError(f"Element with GUID {element_guid} not found in the IFC model.")

    # Find the spatial containment relationship
    spatial_containment = None
    for rel in ifc_model.get_inverse(element):
        if rel.is_a("IfcRelContainedInSpatialStructure"):
            spatial_containment = rel
            break

    if not spatial_containment:
        raise ValueError(f"Element {element_guid} does not have a spatial containment relationship.")

    # Get the current container
    current_container = spatial_containment.RelatingStructure

    # Traverse up the hierarchy to find IfcBuilding or IfcBridge
    parent_container = current_container
    while parent_container and not (parent_container.is_a("IfcBuilding") or parent_container.is_a("IfcBridge")):
        parent_container = next(
            (rel.RelatingObject for rel in ifc_model.get_inverse(parent_container) if rel.is_a("IfcRelAggregates")),
            None,
        )

    if not parent_container:
        raise ValueError(
            f"Could not find an IfcBuilding or IfcBridge in the hierarchy for element {element_guid}."
        )

    # Ensure owner history is defined
    owner_history_list = ifc_model.by_type("IfcOwnerHistory")
    if not owner_history_list:
        raise ValueError("No IfcOwnerHistory found in the model.")
    owner_history = owner_history_list[0]

    # Search for an existing container with the specified name
    existing_container = None
    for rel in ifc_model.get_inverse(parent_container):
        if rel.is_a("IfcRelAggregates"):
            for obj in rel.RelatedObjects:
                if obj.is_a(new_container_class) and obj.Name == bridge_part_name:
                    existing_container = obj
                    break
        if existing_container:
            break

    # Use the existing container or create a new one
    if existing_container:
        new_container = existing_container
    else:
        new_container = ifc_model.create_entity(
            new_container_class,
            GlobalId=ifcopenshell.guid.new(),
            OwnerHistory=owner_history,
            Name=bridge_part_name,
            CompositionType="ELEMENT",
            PredefinedType=bridge_part_type_enum,
        )

        # Link the new container to the IfcBuilding or IfcBridge
        ifc_model.create_entity(
            "IfcRelAggregates",
            GlobalId=ifcopenshell.guid.new(),
            OwnerHistory=owner_history,
            RelatingObject=parent_container,
            RelatedObjects=[new_container],
        )

    # Remove the element from its current spatial container
    related_elements = list(spatial_containment.RelatedElements)  # Convert tuple to list
    if element in related_elements:
        related_elements.remove(element)
    spatial_containment.RelatedElements = tuple(related_elements)  # Reassign updated list as tuple
    if not related_elements:
        # Remove the relationship itself if it becomes empty
        ifc_model.remove(spatial_containment)

    # Reassign the element to the new container
    ifc_model.create_entity(
        "IfcRelContainedInSpatialStructure",
        GlobalId = ifcopenshell.guid.new(),
        OwnerHistory=owner_history,
        RelatingStructure=new_container,
        RelatedElements=[element],
    )

    return new_container

if __name__ == "__main__":
    # Input and output paths
    path = r"C:\00_TECNALIA\01_PROJECTS\TCRINI\00_CODE\RemapBridges\public\ifc_bridges2x3\model02_tablero_2x3_ifc4_ifc4X3.ifc"
    output = r"C:\00_TECNALIA\01_PROJECTS\TCRINI\00_CODE\RemapBridges\public\ifc_bridges_test_class"
    
    base_name = os.path.basename(path)
    name = os.path.splitext(base_name)[0]
    
    output_path = os.path.join(output ,f"{name}_class.ifc")

    # Ensure the output directory exists
    if not os.path.exists(output):
        os.makedirs(output)

    # Open the IFC model
    ifc_model = ifcopenshell.open(path)
    bridge=create_bridge_under_project(ifc_model,bridge_name="MyBridge",bridge_description="PrimaryBridge")
    
    # # Process all IfcWall elements
    # elements = ifc_model.by_type("IfcWall")
    # for element in elements:
    #     element_guid = element.GlobalId
    #     assign_new_container(
    #         ifc_model,
    #         element_guid,
    #         "IfcBridgePart",
    #         bridge_part_name="SUBSTRUCTURE",
    #         bridge_part_type_enum="ABUTMENT",
    #     )

    # Save the updated IFC model
    ifc_model.write(output_path)
    print(f"Updated IFC model saved to {output_path}")