import numpy as np
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d.art3d import Line3DCollection

def visualize_obb_and_plane(points_3d, 
                            corners_obb, 
                            plane_point, 
                            plane_normal, 
                            corners_global_bb = None,
                            plane_size=None, 
                            connect_obb_edges=True, 
                            label_global_bb = True,
                            fig_ax=None):
    """
    Visualize:
      - 'points_3d':        The superstructure cloud (N,3).
      - 'corners_obb':      The 8 OBB corners (8,3).
      - 'plane_point':      A point on the best-fit plane (3,).
      - 'plane_normal':     The plane normal vector (3,).
      - 'plane_size':       Half-size of the plane patch to draw.
                            If None, we auto-compute from corners.
      - 'connect_obb_edges' Whether to draw lines between OBB corners.
      - 'fig_ax':           Optionally supply (fig, ax) to plot into 
                            an existing figure.
    
    Returns:
      fig, ax
    """
    # --------------------------------------------------------------------
    # 1) Prepare figure & axes
    # --------------------------------------------------------------------
    if fig_ax is None:
        fig = plt.figure(figsize=(8, 6))
        ax = fig.add_subplot(111, projection='3d')
    else:
        fig, ax = fig_ax
    
    # --------------------------------------------------------------------
    # 2) Plot the superstructure points
    # --------------------------------------------------------------------
    if points_3d is not None and len(points_3d) > 0:
        ax.scatter(points_3d[:, 0], points_3d[:, 1], points_3d[:, 2],
                   s=2, alpha=0.2, label='Superstructure Points')
    
    # --------------------------------------------------------------------
    # 3) Plot the OBB corners
    # --------------------------------------------------------------------
    if corners_obb is not None and corners_obb.shape == (8, 3):
        ax.scatter(corners_obb[:, 0], corners_obb[:, 1], corners_obb[:, 2],
                   color='r', s=30, label='OBB Corners')
        
        # Optionally connect edges
        if connect_obb_edges:
            # A typical rectangular box has these 12 edges (by corner indices)
            edges = [
                (0,1), (0,2), (1,3), (2,3),   # bottom face
                (4,5), (4,6), (5,7), (6,7),   # top face
                (0,4), (1,5), (2,6), (3,7)    # vertical edges
            ]
            lines = [(corners_obb[i], corners_obb[j]) for (i, j) in edges]
            line_collection = Line3DCollection(lines, colors='gray', linewidths=1)
            ax.add_collection3d(line_collection)

    # --------------------------------------------------------------------
    # 4) Plot the fitted plane as a surface
    # --------------------------------------------------------------------
    if corners_global_bb is not None and corners_global_bb.shape == (8, 3):
        ax.scatter(corners_global_bb[:, 0], corners_global_bb[:, 1], corners_global_bb[:, 2],
                   color='b', s=30, label='Global Bounding Box Corners')
        
        # Optionally connect edges
        if connect_obb_edges:
            edges = [
                (0,1), (0,3), (1,2), (2,3),   # bottom face
                (4,5), (4,7), (5,6), (6,7),   # top face
                (0,4), (1,5), (2,6), (3,7)    # vertical edges
            ]
            lines = [(corners_global_bb[i], corners_global_bb[j]) for (i, j) in edges]
            line_collection = Line3DCollection(lines, colors='blue', linewidths=1)
            ax.add_collection3d(line_collection)

            if label_global_bb:
                for i, (x, y, z) in enumerate(corners_global_bb):
                    ax.text(x, y, z, f'P{i}', color='blue', fontsize=10)

    if plane_point is not None and plane_normal is not None:
        # Ensure normal is a unit vector
        plane_normal = plane_normal / np.linalg.norm(plane_normal)
        
        # Decide plane span (e1, e2) orthonormal to plane_normal
        # e.g., pick an arbitrary vector not parallel to plane_normal:
        arbitrary = np.array([1, 0, 0], dtype=float)
        # If plane_normal is close to [1,0,0], pick something else
        if np.allclose(np.cross(plane_normal, arbitrary), 0, atol=1e-7):
            arbitrary = np.array([0, 1, 0], dtype=float)
        
        e1 = np.cross(plane_normal, arbitrary)
        e1 /= np.linalg.norm(e1)
        e2 = np.cross(plane_normal, e1)
        e2 /= np.linalg.norm(e2)
        
        # If plane_size not given, derive from bounding box extents
        if plane_size is None and corners_obb is not None:
            # A simple approach: bounding box diagonal length
            min_bb = corners_obb.min(axis=0)
            max_bb = corners_obb.max(axis=0)
            diag_len = np.linalg.norm(max_bb - min_bb)
            plane_size = 0.5 * diag_len  # half-diagonal as "radius"
        elif plane_size is None:
            # Fallback default if we don't have corners
            plane_size = 10.0
        
        # Create a mesh grid in local (e1,e2) space
        num_steps = 10
        u_vals = np.linspace(-plane_size, plane_size, num_steps)
        v_vals = np.linspace(-plane_size, plane_size, num_steps)
        plane_X = []
        plane_Y = []
        plane_Z = []
        
        for u in u_vals:
            row_x = []
            row_y = []
            row_z = []
            for v in v_vals:
                # param point = plane_point + u * e1 + v * e2
                p = plane_point + u*e1 + v*e2
                row_x.append(p[0])
                row_y.append(p[1])
                row_z.append(p[2])
            plane_X.append(row_x)
            plane_Y.append(row_y)
            plane_Z.append(row_z)
        
        plane_X = np.array(plane_X)
        plane_Y = np.array(plane_Y)
        plane_Z = np.array(plane_Z)
        
        ax.plot_surface(plane_X, plane_Y, plane_Z, 
                        alpha=0.2, color='cyan', label='Fitted Plane')
    
    # --------------------------------------------------------------------
    # 5) Make it look nice
    # --------------------------------------------------------------------
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    ax.legend(loc='best')
    
    # Optional: set equal aspect so the scaling is the same in all directions
    # This function is a common trick for matplotlib 3D
    set_axes_equal(ax)
    
    plt.tight_layout()
    return fig, ax

def set_axes_equal(ax: plt.Axes):
    """
    Helper to make the 3D plot have equal scale on all axes.
    """
    limits = np.array([ax.get_xlim3d(), ax.get_ylim3d(), ax.get_zlim3d()])
    spans = limits[:,1] - limits[:,0]
    max_span = max(spans)
    mid = np.mean(limits, axis=1)
    ax.set_xlim3d([mid[0] - 0.5*max_span, mid[0] + 0.5*max_span])
    ax.set_ylim3d([mid[1] - 0.5*max_span, mid[1] + 0.5*max_span])
    ax.set_zlim3d([mid[2] - 0.5*max_span, mid[2] + 0.5*max_span])
